/* ser_serial.c
 * 
 * Provides cross-platform access to the serial port.
 * 
 * Copyright (C) 2000, Winford Engineering.
 * See file LICENSE.txt for details.
 * 
 * Winford Engineering
 * www.winfordeng.com
 */


/* 
 * These functions are specialized for this particular
 * library, and are not meant to be general purpose.
 * These functions will NOT be used directly by the user.
 * 
 * MUST have either LINUX or WIN95 or WINNT #defined.
 */
#include "ser_serial.h"

static int ser_serial_check_baud(unsigned int baud)
{ /* This is an internal function used to check if a baud
   * rate specified is valid.
   * 
   * This same function is used for both Linux and Windows,
   * so it is above the COMMONWIN and LINUX sections below.
   * 
   * Returns 0 for valid and -1 for invalid
   */

	switch(baud)
	{
	case SER_1200:
	case SER_2400:
	case SER_4800:
	case SER_9600:
	case SER_19200:
	case SER_38400:
		return(0);
		break;
	}

	return(-1);
}

#ifdef COMMONWIN
#include <windows.h>
	
int ser_serial_open(ser_serial* serial, char* path, unsigned int baud)
{
	/* this function opens the serial port specified by "path"  path
	* will vary depending on the operating system, where in linux
	* it is the path to the serial device file used, and in Windows
	* it is simply the name of the COM port.  This function also
	* sets the baud rate of the serial port. This library currently
	* supports baud rates doubling from 1200 up to 38400, while 
	* certain serial board models will likely limit to just a
	* few of these choices.  This is up to each device's API.
	* 
	* Return Values: 0 for success
	*                SER_SERIAL_ERROR if failed to open serial port
	*                
	*/

	HANDLE hserial;
	DCB *newattr;
	COMMTIMEOUTS timeouts;

	if(ser_serial_check_baud(baud))
		return(SER_INVALID_BAUD);

	/* Windows is really stupid here.  We have to open the serial port 
	 * for OVERLAPPED file access here, because you can't be writing 
	 * to the serial port with one thread and reading it from another.
	 * Actually, you can get away with it on Win98, but it's out of 
	 * the question on NT or 2000.  That's why we have to open with 
	 * OVERLAPPED flag, otherwise things would be a good bit easier.
	 */ 
	
	if( (hserial=CreateFile(path, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING, NULL)) == INVALID_HANDLE_VALUE )
		return(SER_SERIAL_ERROR);

	serial->fd = hserial;
	
	(serial->oldattr).DCBlength = sizeof(serial->oldattr);
	
	if( GetCommState(hserial, &(serial->oldattr)) == 0 )
		goto error_1;

	newattr = &(serial->newattr);
	*newattr = serial->oldattr;

	switch(baud)
	{
		case SER_1200:
			newattr->BaudRate = CBR_1200;
			break;
		case SER_2400:
			newattr->BaudRate = CBR_2400;
			break;
		case SER_4800:
			newattr->BaudRate = CBR_4800;
			break;
		case SER_9600:
			newattr->BaudRate = CBR_9600;
			break;
		case SER_19200:
			newattr->BaudRate = CBR_19200;
			break;
		case SER_38400:
			newattr->BaudRate = CBR_38400;
			break;
	}

	newattr->fParity = FALSE;
	newattr->fOutxCtsFlow = FALSE;
	newattr->fOutxDsrFlow = FALSE;
	newattr->fDtrControl = DTR_CONTROL_DISABLE;
	newattr->fDsrSensitivity = FALSE;
	newattr->fOutX = FALSE;
	newattr->fInX = FALSE;
	newattr->fNull = FALSE;
	newattr->fRtsControl = RTS_CONTROL_DISABLE;
	newattr->fAbortOnError = FALSE;
	newattr->ByteSize = 8;
	newattr->Parity = NOPARITY;
	newattr->StopBits = ONESTOPBIT;
	
	if( SetCommState(hserial, newattr) == 0)
		goto error_1;

	/* According to Microsoft documentation, the following setup will cause ReadFile
	 * to return immediately if a character is available, and wait SER_SERIAL_TIMEOUT
	 * milliseconds for a character arrive if none is available.  After that, if no 
	 * character has arrived it will timeout
	 */
	timeouts.ReadIntervalTimeout = MAXDWORD;
	timeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
	timeouts.ReadTotalTimeoutConstant = SER_SERIAL_TIMEOUT; /* set default timeout period */
	serial->timeout = SER_SERIAL_TIMEOUT; /* set default timeout period */

	/* Don't use any write timeouts.  They shouldn't be applicable anyways because
	 * we don't use any flow control.
	 */
	timeouts.WriteTotalTimeoutMultiplier = 0;
	timeouts.WriteTotalTimeoutConstant = 0;

	if( SetCommTimeouts(hserial, &timeouts) == 0 )
		goto error_1;

	/* create new manual reset events... initialize overlapped stuff */
	if( ((serial->overwrite).hEvent = CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL )
		goto error_1;
	(serial->overwrite).Offset = 0;
	(serial->overwrite).OffsetHigh = 0;
	

	if( ((serial->overread).hEvent = CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL )
		goto error_2;
	(serial->overread).Offset = 0;
	(serial->overread).OffsetHigh = 0;
	
	return(0);
error_2: /* the OVERLAPPED write event must be closed */
	CloseHandle((serial->overwrite).hEvent);
	
error_1: /* something happened, so the handle to the serial port must be closed */
	CloseHandle(hserial);
	return(SER_SERIAL_ERROR);
}

int ser_serial_timeout(ser_serial *serial, unsigned int timeout)
{
/* this function modifies the timeout time for the serial line
 * timeout is in milliseconds
 */
	COMMTIMEOUTS comto;

	if( GetCommTimeouts(serial->fd, &comto) == 0 )
		return(SER_SERIAL_ERROR);

	comto.ReadTotalTimeoutConstant = timeout;

	/* store value in our own structure too */
	serial->timeout = timeout;

	if( SetCommTimeouts(serial->fd, &comto) == 0 )
		return(SER_SERIAL_ERROR);

	return(0);
}

int ser_serial_write(ser_serial *serial, char* data, int length)
{
	/* writes data to the serial port.
	 * Return Value: number of bytes written
	 * 				  or negative error values.
	 */
	DWORD written;

	if(length == 0)
		return(-1);
		
	if( WriteFile(serial->fd, data, length, &written, &(serial->overwrite)) == 0 ) /* either error or asynchronous */
	{
		if(GetLastError() != ERROR_IO_PENDING)
			return(-1);
		else
			GetOverlappedResult(serial->fd, &(serial->overwrite), &written, TRUE); /* wait for completion */
	}
	/* If the function didn't fail, simply return the number written */			
		
	return(written);
}

int ser_serial_readline(ser_serial *serial, char *data, int length, int *actual)
{
	/* Reads a line (or up till length) of characters from the
	 * serial port.  
	 * 
	 * "length" is the total length in bytes of the buffer
	 * specified by data.  The function will null terminate
	 * results, so allocate and pass a buffer one byte larger
	 * than the data that you need. 
	 * 
	 * Return Value: 0 for success, full line or full buffer; length returned by "actual".
	 * 				 1 for timeout, length read so far returned by "actual"
	 * 				 -1 for general error: can't read, etc.
	 */

	/* NOTE: This function reads the serial port using ReadFile.
	 * If a read times out without reading a character, ReadFile returns a nonzero
	 * number, indicating success, but number of bytes read will be zero.
	 */
	 
	int numread=0; /* number of bytes read and put into the buffer */
	DWORD num; /* number of bytes read by an individual ReadFile function */
	BOOL retval;

	while( (length-numread - 1) > 0 ) /* -1 to allow for null termination later */
	{
		num = 0; /* in case ReadFile doesn't modify num, on an error say, set it for the benifit of checks made below */

		if( (retval=ReadFile(serial->fd, data + numread, 1, &num, &(serial->overread))) == 0 )
		{
			if(GetLastError() != ERROR_IO_PENDING)
				break;
			else
			{               /* No need to worry about timeouts in here.  The serial port timeout will take effect */
				if( (retval=GetOverlappedResult(serial->fd, &(serial->overread), &num, TRUE)) == 0)
					break;
			}
		}

		if( num <= 0 ) /* if no characters were actually read */
			break;
			
		if( (data[numread] == '\r') || (data[numread] == '\n') )
			break;
		else
			numread++;
	}
	
	if(numread)
	{
		data[numread] = '\0';
		*actual = numread;
		if( num > 0 ) /* If the last read obtained a character, it must have been a newline
		               * because we are here. so even if the read call failed, return what we've got so far 
		               */
			return(0);
		else
			return(1);/* indicate timeout since last read didn't get a character */
	}
	else
	{
		*actual = 0;
		if( retval && num > 0) /* There's no failure, we received a character, but don't have any data.  Must have only received a newline */
		{
			data[0] = '\0';
			return(0);
		}
		else if(retval == 0)
			return(-1);
		else /* no error, but haven't received any characters. timeout */
			return(1);
	}
}


int ser_serial_close(ser_serial *serial)
{
	/* closes the serial connection. restores the original
	 * state of the serial line and settings.
	 * Return Value: 0 for success
	 *               -1 for error
	 */

	BOOL retval; /* indicates whether either function failed */

	/* restore the old settings of the serial port */
	retval =  !(SetCommState(serial->fd, &(serial->oldattr)));
	/* close the port */
	retval |= !(CloseHandle(serial->fd));

	retval |= !(CloseHandle((serial->overwrite).hEvent));

	retval |= !(CloseHandle((serial->overread).hEvent));
	
	if(retval)
		return(-1);
	else
		return(0);	
}

#endif 
/* end of COMMONWIN */

#ifdef LINUX
#include <unistd.h>
#include <stdio.h>

int ser_serial_open(ser_serial* serial, char* path, unsigned int baud)
{
	/* this function opens the serial port specified by "path"  path
	* will vary depending on the operating system, where in linux
	* it is the path to the serial device file used, and in Windows
	* it is simply the name of the COM port.  This function also
	* sets the baud rate of the serial port. This library currently
	* supports baud rates doubling from 1200 up to 38400, while 
	* certain serial board models will likely limit to just a
	* few of these choices.  This is up to each device's API.
	* 
	* Return Values: 0 for success
	*                SER_SERIAL_ERROR if failed to open serial port
	*                
	*/

	struct termios attr;

	if(ser_serial_check_baud(baud))
		return(SER_INVALID_BAUD);
	
	if((serial->fd = open(path, O_RDWR | O_NOCTTY)) < 0)
		return(SER_SERIAL_ERROR);

	if(tcgetattr(serial->fd, &attr))
	{
		close(serial->fd);
		return(SER_SERIAL_ERROR);
	}

	serial->oldattr = attr; /* save old attributes for restoration during close call, etc. */

	attr.c_iflag = 0;
	attr.c_oflag = 0;
	attr.c_cflag = 0;

	switch(baud)
	{
		case SER_1200:
			cfsetispeed(&attr, B1200);
			cfsetospeed(&attr, B1200);
			break;
		case SER_2400:
			cfsetispeed(&attr, B2400);
			cfsetospeed(&attr, B2400);
			break;
		case SER_4800:
			cfsetispeed(&attr, B4800);
			cfsetospeed(&attr, B4800);
			break;
		case SER_9600:
			cfsetispeed(&attr, B9600);
			cfsetospeed(&attr, B9600);
			break;
		case SER_19200:
			cfsetispeed(&attr, B19200);
			cfsetospeed(&attr, B19200);
			break;
		case SER_38400:
			cfsetispeed(&attr, B38400);
			cfsetospeed(&attr, B38400);
			break;
	}

	attr.c_cflag |= (CREAD | CS8 | CLOCAL); /* set flags */
	
	attr.c_lflag = 0; /* non-canonical, no echo, etc. */

	attr.c_cc[VMIN] = 0;
	attr.c_cc[VTIME] = SER_SERIAL_TIMEOUT / 100; /* convert msec to tenths of a second */

	serial->newattr = attr;
	
	if( tcsetattr(serial->fd, TCSANOW, &attr) )
	{/* if an error occurs */
		close(serial->fd);
		return(SER_SERIAL_ERROR);
	}
	
	return(0);
}

int ser_serial_timeout(ser_serial * serial, unsigned int timeout)
{
/* this function modifies the timeout time for the serial line
 * timeout is in milliseconds
 */
	serial->newattr.c_cc[VTIME] = timeout / 100;

	if( tcsetattr(serial->fd, TCSADRAIN, &(serial->newattr)) )
		return(SER_SERIAL_ERROR);

	return(0);
}

int ser_serial_write(ser_serial* serial, char* data, int length)
{
	/* writes data to the serial port.
	 * Return Value: number of bytes written
	 * 				  or negative error values.
	 */
	int retval;

	if(length == 0)
		return(-1);
		
	retval = write(serial->fd, data, length);
	
	if(retval > 0)
		return(retval);
	else
		return(-1);
}

int ser_serial_readline(ser_serial* serial, char* data, int length, int* actual)
{
	/* Reads a line (or up till length) of characters from the
	 * serial port.  
	 * 
	 * "length" is the total length in bytes of the buffer
	 * specified by data.  The function will null terminate
	 * results, so allocate and pass a buffer one byte larger
	 * than the data that you need. 
	 * 
	 * Return Value: 0 for success, full line or full buffer; length returned by "actual".
	 * 				 1 for timeout, length read so far returned by "actual"
	 * 				 -1 for general error: can't read, etc.
	 */
	int numread=0;
	int retval=0;

	while( (length-numread - 1) > 0 )/* -1 to allow for null termination later */
	{
		if( (retval=read(serial->fd, data + numread, 1)) > 0 )
		{/* we've got some valid input */
			if( (data[numread] == '\r') || (data[numread] == '\n') )
					break;
			else
				numread++;
		}
		else
			break;
	}
	
	if(numread)
	{
		data[numread] = '\0';
		*actual = numread;
		if(retval)/* even if the read call failed, return what we've got so far */
			return(0);
		else
			return(1);/* indicate timeout if last read didn't get a character */
	}
	else
	{
		*actual = 0;
		if(retval == 1)/* all we got is a newline */
		{
			data[0] = '\0';
			return(0);
		}
		else if(retval)
			return(-1);
		else
			return(1);
	}
}

int ser_serial_close(ser_serial * serial)
{
	/* closes the serial connection. restores the original
	 * state of the serial line and settings.
	 * Return Value: 0 for success
	 *               -1 for error
	 */

	int retval;

	retval = tcsetattr(serial->fd, TCSADRAIN, &(serial->oldattr));
	retval |= close(serial->fd);

	if(retval)
		return(-1);
	else
		return(0);	
	 
}
#endif 
/* end of LINUX */