/* ser_threads.h
 * 
 * Provides cross-platform thread creation/control and synchronization 
 * methods needed by this serial API.
 * 
 * Copyright (C) 2000, Winford Engineering.
 * See file LICENSE.txt for details.
 * 
 * Winford Engineering
 * www.winfordeng.com
 */

/* NOTE:
 * To compile you MUST have either LINUX or WIN95 or WINNT #defined.
 */
 
#ifndef ser_threads_h
#define ser_threads_h

#ifdef WIN95
#define COMMONWIN
#endif
#ifdef WINNT
#define COMMONWIN
#endif

#ifdef COMMONWIN
#include <windows.h>
typedef struct
{
	HANDLE mutex_pred; /* mutex locked by "prewait" which allows the user to inspect the predicate */
	HANDLE event_sig;  /* Manual-reset event providing the signal for the event.  This stays signalled until the last waiting thread sees it.  That thread then resets it */
	HANDLE event_done; /* Manual-reset event signalling that all the waiting threads have "seen" the event's signal */
	int waiting; /* number of threads waiting on this event */
	HANDLE mutex_waiting; /* protects access to the waiting variable */

} ser_event;

typedef HANDLE ser_mutex;

/* LPTHREAD_START_ROUTINE has this format:
   DWORD WINAPI funcname(LPVOID);
 */
typedef LPTHREAD_START_ROUTINE ser_thread;

#endif

#ifdef LINUX
#include <pthread.h>
typedef struct 
{
	pthread_mutex_t mutex;
	pthread_cond_t cond;
} ser_event;
typedef void *(*ser_thread)(void *);

typedef struct
{
	pthread_mutex_t mutex;
	pthread_cond_t cond;
	int busy;
} ser_mutex;
#endif

int ser_thread_create(ser_thread func, void* arg);


int ser_event_init(ser_event* event);
int ser_event_prewait(ser_event* event);
/* Obtains the mutex so the predicate can be examined.
 * either the release or the wait function should be quickly
 * called. 
 */
int ser_event_wait(ser_event* event, unsigned int* timeout);
/* after wait returns, the thread has exclusive access to the
 * data being protected by the event.  The thread should
 * call the release function as soon as possible
 * 
 * timeout specifies the number of milliseconds to wait for the event
 * before timing out.  If a NULL pointer is passed, no timeout is used.
 * If the event is signalled before the timeout, the number of remaining
 * milliseconds before the scheduled timeout is written into timeout.
 * 
 * Return Values: 0 for success
 * 				  1 for timeout
 * 				  -1 for error
 */

int ser_event_broadcast(ser_event* event);
/* The event should be owned by the calling thread (either by prewait
 *  or after wait)
 * This function releases any threads waiting on the
 * event.  However, those threads cannot execute unless
 * the release function is called after this function
 * is called.
 */

int ser_event_release(ser_event* event);
/* This function releases the thread's hold on the predicate 
 * condition.  It does NOT signam the event.
 * the signal function should be called immediately before
 * this function if desired
 */

int ser_event_destroy(ser_event* event);

int ser_mutex_init(ser_mutex* mutex);
int ser_mutex_wait(ser_mutex* mutex, unsigned int* timeout);
/* timeout specifies the number of milliseconds to wait for the mutex
 * before timing out.  If a NULL pointer is passed, no timeout is used.
 * If the mutex is obtained before the timeout, the number of remaining
 * milliseconds before the scheduled timeout is written into timeout.
 * 
 * Return Values: 0 for success (got mutex)
 * 				  1 for timeout
 * 				  -1 for error
 */

int ser_mutex_release(ser_mutex* mutex);
int ser_mutex_destroy(ser_mutex* mutex);

#endif // ser_threads_h
